@@ -2,7 +2,10 @@
 
 #include "ftpd.h"
 #include "pure-pw.h"
+#include "safe_rw.h"
+#include "alt_arc4random.h"
 #include "../puredb/src/puredb_write.h"
+#include "utils.h"
 #ifdef HAVE_POSIX_TERMIOS
 # include <termios.h>
 #elif defined(HAVE_TERMIO_H)
@@ -17,6 +20,10 @@
 # include <getopt.h>
 #endif
 
+#ifdef HAVE_LIBSODIUM
+# include <sodium.h>
+#endif
+
 #ifdef WITH_DMALLOC
 # include <dmalloc.h>
 #endif
@@ -32,22 +39,23 @@ static void disable_echo(void)
     }
 #ifdef ECHO
 # ifdef HAVE_POSIX_TERMIOS
-    {    
+    {
         struct termios p;
-        
+
         if (tcgetattr(0, &p) != 0) {
             return;
         }
         p.c_lflag &= ~ECHO;
-#  ifndef TCSAFLUSH
-#   define TCSAFLUSH 0
-#  endif
+#  if defined(TCSASOFT) && defined(TCSAFLUSH)
+        tcsetattr(0, TCSASOFT | TCSAFLUSH, &p);
+#  elif defined(TCSAFLUSH)
         tcsetattr(0, TCSAFLUSH, &p);
+#  endif
     }
 # elif defined(HAVE_TERMIO_H) && defined(TCGETA)
     {
         struct termio tty;
-        
+
         if (ioctl(0, TCGETA, &tty) != 0) {
             return;
         }
@@ -57,7 +65,7 @@ static void disable_echo(void)
 # else
     {
         struct sgttyb tty;
-        
+
         if (ioctl(0, TIOCGETP, &tty) != 0) {
             return;
         }
@@ -72,25 +80,27 @@ static void enable_echo(void)
 {
     if (!isatty(0)) {
         return;
-    }    
+    }
 #ifdef ECHO
 # ifdef HAVE_POSIX_TERMIOS
     {
         struct termios p;
-        
+
         if (tcgetattr(0, &p) != 0) {
             return;
         }
         p.c_lflag |= ECHO;
-#  ifndef TCSAFLUSH
-#   define TCSAFLUSH 0
+#  if defined(TCSASOFT) && defined(TCSAFLUSH)
+        tcsetattr(0, TCSASOFT | TCSAFLUSH, &p);
+#  elif defined(TCSAFLUSH)
+        tcsetattr(0, TCSAFLUSH, &p);
 #  endif
         tcsetattr(0, TCSAFLUSH, &p);
     }
 # elif defined(HAVE_TERMIO_H) && defined(TCGETA)
     {
         struct termio tty;
-        
+
         if (ioctl(0, TCGETA, &tty) != 0) {
             return;
         }
@@ -100,7 +110,7 @@ static void enable_echo(void)
 # else
     {
         struct sgttyb tty;
-        
+
         if (ioctl(0, TIOCGETP, &tty) != 0) {
             return;
         }
@@ -122,11 +132,11 @@ static char *my_strtok2(char *str, const char delim)
 {
     static char *s;
     static char save;
-    
+
     if (str != NULL) {
         if (*str == 0) {
             return NULL;
-        }        
+        }
         s = str;
         scan:
         while (*s != 0 && *s != delim) {
@@ -134,15 +144,15 @@ static char *my_strtok2(char *str, const char delim)
         }
         save = *s;
         *s = 0;
-        
+
         return str;
     }
-    if (s == NULL || save == 0) {        
+    if (s == NULL || save == 0) {
         return NULL;
     }
     s++;
     str = s;
-    
+
     goto scan;
 }
 
@@ -169,7 +179,7 @@ static void help(void)
          "                [-q <upload ratio>] [-Q <download ratio>]\n"
          "                [-r <allow client ip>/<mask>] [-R <deny client ip>/<mask>]\n"
          "                [-i <allow local ip>/<mask>] [-I <deny local ip>/<mask>]\n"
-     "                [-y <max number of concurrent sessions>]\n"
+         "                [-y <max number of concurrent sessions>]\n"
          "                [-z <hhmm>-<hhmm>] [-m]\n"
          "\n"
          "pure-pw usermod <login> -f <passwd file> -u <uid> [-g <gid>]\n"
@@ -179,7 +189,7 @@ static void help(void)
          "                [-q <upload ratio>] [-Q <download ratio>]\n"
          "                [-r <allow client ip>/<mask>] [-R <deny client ip>/<mask>]\n"
          "                [-i <allow local ip>/<mask>] [-I <deny local ip>/<mask>]\n"
-     "                [-y <max number of concurrent sessions>]\n"     
+         "                [-y <max number of concurrent sessions>]\n"
          "                [-z <hhmm>-<hhmm>] [-m]\n"
          "\n"
          "pure-pw userdel <login> [-f <passwd file>] [-m]\n"
@@ -189,12 +199,13 @@ static void help(void)
          "pure-pw show    <login> [-f <passwd file>]\n"
          "\n"
          "pure-pw mkdb    [<puredb database file> [-f <passwd file>]]\n"
+         "                [-F <puredb file>]\n"
          "\n"
          "pure-pw list    [-f <passwd file>]\n"
          "\n"
          "-d <home directory> : chroot user (recommended)\n"
          "-D <home directory> : don't chroot user\n"
-     "-<option> '' : set this option to unlimited\n"
+         "-<option> '' : set this option to unlimited\n"
          "-m : also update the " DEFAULT_PW_DB " database\n"
          "For a 1:10 ratio, use -q 1 -Q 10\n"
          "To allow access only between 9 am and 6 pm, use -z 0900-1800\n"
@@ -202,7 +213,7 @@ static void help(void)
 #ifndef WITH_PUREDB
     puts("*WARNING* : that pure-ftpd server hasn't been compiled with puredb support\n");
 #endif
-    exit(EXIT_SUCCESS);   
+    exit(EXIT_SUCCESS);
 }
 
 static void no_mem(void)
@@ -211,137 +222,95 @@ static void no_mem(void)
     exit(EXIT_FAILURE);
 }
 
-#ifdef PROBE_RANDOM_AT_RUNTIME
-static void pw_zrand_probe(void)
-{
-    static const char * const devices[] = {
-        "/dev/arandom", "/dev/urandom", "/dev/random", NULL
-    };
-    const char * const *device = devices;
-    
-    do {
-        if (access(*device, F_OK | R_OK) == 0) {
-            random_device = *device;
-            break;
-        }
-        device++;
-    } while (*device != NULL);
-}
-#endif
-
-static unsigned int pw_zrand(void)
-{
-    int fd;
-    int ret;
-    
-    if (
-#ifdef PROBE_RANDOM_AT_RUNTIME
-        ((fd = open(random_device, O_RDONLY | O_NONBLOCK)) == -1)
-#elif defined(HAVE_DEV_ARANDOM)
-        ((fd = open("/dev/arandom", O_RDONLY | O_NONBLOCK)) == -1)
-#elif defined(HAVE_DEV_URANDOM)
-        ((fd = open("/dev/urandom", O_RDONLY | O_NONBLOCK)) == -1)
-#else
-        ((fd = open("/dev/random", O_RDONLY | O_NONBLOCK)) == -1)
-#endif        
-        ) {
-        nax:
-#ifdef HAVE_ARC4RANDOM
-        return (unsigned int) arc4random();
-#elif defined HAVE_RANDOM
-        return (unsigned int) random();
-#else
-        return (unsigned int) rand();
-#endif
-    }
-    if (read(fd, &ret, sizeof ret) != (ssize_t) sizeof ret) {
-        close(fd);
-        goto nax;
-    }
-    close(fd);    
-    
-    return (unsigned int) ret;
-}
-
 static char *best_crypt(const char * const pwd)
 {
+#ifdef HAVE_LIBSODIUM
+    static char hash[crypto_pwhash_scryptsalsa208sha256_STRBYTES];
+
+    if (crypto_pwhash_scryptsalsa208sha256_str
+        (hash, pwd, strlen(pwd),
+            crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE,
+            crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE) != 0) {
+        no_mem();
+    }
+    return hash;
+#else
     static const char crcars[64] =
         "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
     const char *crypted;
-    
-    if ((crypted = (const char *)      /* Blowfish */
-         crypt("test", "$2a$07$1234567890123456789012")) != NULL &&        
-        strcmp(crypted, "$2a$07$123456789012345678901uKO4"
-               "/IReKqBzRzT6YaajGvw20UBdHW7m") == 0) {
-        char salt[] = "$2a$07$0000000000000000000000";        
+
+    if ((crypted = (const char *)      /* bcrypt */
+         crypt("test", "$2a$08$1234567890123456789012")) != NULL &&
+        strcmp(crypted,
+               "$2a$08$123456789012345678901uBdmsfIXjJcWQwz1wT/IZrWhimJ6xy6a")
+        == 0) {
+        char salt[] = "$2a$08$0000000000000000000000";
         int c = 28;
-        
-        do {            
+
+        do {
             c--;
-            salt[c] = crcars[pw_zrand() & 63];
+            salt[c] = crcars[alt_arc4random() & 63];
         } while (c > 7);
-        
-        return (char *) crypt(pwd, salt);        
+
+        return (char *) crypt(pwd, salt);
+    } else if ((crypted = (const char *)    /* SHA-512 */
+                crypt("test", "$6$1234567890123456$")) != NULL &&
+               strcmp(crypted,
+                      "$6$1234567890123456$d.pgKQFaiD8bRiExg5NesbGR/"
+                      "3u51YvxeYaQXPzx4C6oSYREw8VoReiuYZjx0V9OhGVTZF"
+                      "qhc6emAxT1RC5BV.") == 0) {
+        char salt[] = "$6$0000000000000000";
+        int c = 18;
+
+        do {
+            c--;
+            salt[c] = crcars[alt_arc4random() & 63];
+        } while (c > 3);
+
+        return (char *) crypt(pwd, salt);
     } else if ((crypted = (const char *)    /* MD5 */
                 crypt("test", "$1$12345678$")) != NULL &&
                strcmp(crypted, "$1$12345678$oEitTZYQtRHfNGmsFvTBA/") == 0) {
         char salt[] = "$1$00000000";
         int c = 10;
-        
-        do {            
-            c--;
-            salt[c] = crcars[pw_zrand() & 63];
-        } while (c > 3);
-        
-        return (char *) crypt(pwd, salt);
-    } else if ((crypted = (const char *)    /* Extended DES */
-                crypt("test", "_.../1234")) != NULL &&
-               strcmp(crypted, "_.../1234PAPUVmqGzpU") == 0) {
-        char salt[] = "_.../0000";
-        int c = 8;
-        
+
         do {
             c--;
-            salt[c] = crcars[pw_zrand() & 63];
-        } while (c > 5);
-        
+            salt[c] = crcars[alt_arc4random() & 63];
+        } while (c > 3);
+
         return (char *) crypt(pwd, salt);
+    } else {
+        fprintf(stderr, "No useable password hashing function found\n");
+        exit(EXIT_FAILURE);
     }
-    /* Simple DES */
-    {
-        char salt[] = "00";
-        
-        salt[0] = crcars[pw_zrand() & 63];
-        salt[1] = crcars[pw_zrand() & 63];
-        
-        return (char *) crypt(pwd, salt);        
-    }    
+#endif
 }
 
 char *newpasswd_filename(const char * const file)
 {
     size_t sizeof_file2;
     char *file2;
-    
+
     sizeof_file2 = strlen(file) + sizeof NEWPASSWD_SUFFIX;
     if ((file2 = malloc(sizeof_file2)) == NULL) {
         return NULL;
     }
     (void) snprintf(file2, sizeof_file2, "%s%s", file, NEWPASSWD_SUFFIX);
-    
+
     return file2;
 }
 
 static void strip_lf(char *str)
 {
     char *f;
-    
+
     if (str == NULL) {
         return;
     }
     if ((f = strchr(str, '\r')) != NULL) {
         *f = 0;
-    }    
+    }
     if ((f = strchr(str, '\n')) != NULL) {
         *f = 0;
     }
@@ -354,7 +323,7 @@ static int parse_pw_line(char *line, PWInfo * const pwinfo)
     pwinfo->gecos = NULL;
     pwinfo->home = NULL;
     pwinfo->allow_local_ip = pwinfo->deny_local_ip = NULL;
-    pwinfo->allow_client_ip = pwinfo->deny_client_ip = NULL;    
+    pwinfo->allow_client_ip = pwinfo->deny_client_ip = NULL;
     pwinfo->has_bw_dl = 0;
     pwinfo->has_bw_ul = 0;
     pwinfo->has_quota_files = 0;
@@ -367,7 +336,7 @@ static int parse_pw_line(char *line, PWInfo * const pwinfo)
     pwinfo->gid = (gid_t) 0;
     pwinfo->has_per_user_max = 0;
     pwinfo->per_user_max = 0U;
-    
+
     if ((line = my_strtok2(line, *PW_LINE_SEP)) == NULL || *line == 0) {   /* account */
         return -1;
     }
@@ -397,7 +366,7 @@ static int parse_pw_line(char *line, PWInfo * const pwinfo)
     if (*line != '/') {
         return -1;
     }
-    pwinfo->home = line;    
+    pwinfo->home = line;
     if ((line = my_strtok2(NULL, *PW_LINE_SEP)) == NULL) {   /* bw_ul */
         return 0;
     }
@@ -407,7 +376,7 @@ static int parse_pw_line(char *line, PWInfo * const pwinfo)
     }
     if ((line = my_strtok2(NULL, *PW_LINE_SEP)) == NULL) {   /* bw_dl */
         return 0;
-    } 
+    }
     if (*line != 0) {
         pwinfo->has_bw_dl = 1;
         pwinfo->bw_dl = strtoul(line, NULL, 10);
@@ -418,7 +387,7 @@ static int parse_pw_line(char *line, PWInfo * const pwinfo)
     if (*line != 0) {
         pwinfo->ul_ratio = (unsigned int) strtoul(line, NULL, 10);
         if (pwinfo->ul_ratio > 0U) {
-            pwinfo->has_ul_ratio = 1;     
+            pwinfo->has_ul_ratio = 1;
         }
     }
     if ((line = my_strtok2(NULL, *PW_LINE_SEP)) == NULL) {   /* ratio down */
@@ -477,7 +446,7 @@ static int parse_pw_line(char *line, PWInfo * const pwinfo)
         pwinfo->time_end < 2360 && (pwinfo->time_end % 100) < 60) {
         pwinfo->has_time = 1;
     }
-    
+
     return 0;
 }
 
@@ -487,7 +456,7 @@ static int fetch_pw_account(const char * const file, PWInfo * const pwinfo,
 {
     FILE *fp;
     int ret = -1;
-    
+
     if (file == NULL || pwinfo == NULL || line == NULL ||
         sizeof_line < (size_t) 2U || login == NULL ||
         *login == 0) {
@@ -530,13 +499,13 @@ static FILE *create_newpasswd(const char * const file,
     int found = 0;
     size_t skip_login_len;
     char line[LINE_MAX];
-    
-    fp = fopen(file, "r");    
+
+    fp = fopen(file, "r");
     if ((fd2 = open(file2, O_EXCL | O_NOFOLLOW |
                     O_CREAT | O_WRONLY, (mode_t) 0700)) == -1) {
         if (fp != NULL) {
             fclose(fp);
-        }        
+        }
         return NULL;
     }
     if ((fp2 = fdopen(fd2, "w")) == NULL) {
@@ -544,7 +513,7 @@ static FILE *create_newpasswd(const char * const file,
             fclose(fp);
         }
         close(fd2);
-        
+
         return NULL;
     }
     if (fp != NULL) {
@@ -566,11 +535,11 @@ static FILE *create_newpasswd(const char * const file,
             }
             if (fputs(line, fp2) < 0) {
                 err:
-                fclose(fp2);                
+                fclose(fp2);
                 unlink(file2);
                 if (fp != NULL) {
                     fclose(fp);
-                }                                                
+                }
                 return NULL;
             }
         }
@@ -583,7 +552,7 @@ static FILE *create_newpasswd(const char * const file,
     if (fp != NULL) {
         fclose(fp);
     }
-    
+
     return fp2;
 }
 
@@ -592,12 +561,12 @@ static int add_new_pw_line(FILE * const fp2, const PWInfo * const pwinfo)
     if (fp2 == NULL) {
         return -1;
     }
-    if (fprintf(fp2, 
+    if (fprintf(fp2,
                 "%s" PW_LINE_SEP           /* account */
                 "%s" PW_LINE_SEP           /* password */
                 "%lu" PW_LINE_SEP          /* uid */
                 "%lu" PW_LINE_SEP          /* gid */
-                "%s" PW_LINE_SEP           /* gecos */                
+                "%s" PW_LINE_SEP           /* gecos */
                 "%s" PW_LINE_SEP           /* home */
                 , pwinfo->login, pwinfo->pwd,
                 (unsigned long) pwinfo->uid, (unsigned long) pwinfo->gid,
@@ -610,7 +579,7 @@ static int add_new_pw_line(FILE * const fp2, const PWInfo * const pwinfo)
             return -1;
         }
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
     if (pwinfo->has_bw_dl != 0) {
@@ -619,25 +588,25 @@ static int add_new_pw_line(FILE * const fp2, const PWInfo * const pwinfo)
             return -1;
         }
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
     if (pwinfo->has_ul_ratio != 0) {
         if (fprintf(fp2, "%u",             /* ratio up */
                     pwinfo->ul_ratio) < 0) {
             return -1;
-        }        
+        }
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
     if (pwinfo->has_dl_ratio != 0) {
         if (fprintf(fp2, "%u",             /* ratio down */
                     pwinfo->dl_ratio) < 0) {
             return -1;
-        }        
+        }
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
     if (pwinfo->has_per_user_max != 0) {
@@ -645,49 +614,49 @@ static int add_new_pw_line(FILE * const fp2, const PWInfo * const pwinfo)
         return -1;
     }
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
-    }    
+    }
     if (pwinfo->has_quota_files != 0) {
         if (fprintf(fp2, "%llu",           /* files quota */
                     pwinfo->quota_files) < 0) {
             return -1;
         }
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
-    if (pwinfo->has_quota_size != 0) {    
+    if (pwinfo->has_quota_size != 0) {
         if (fprintf(fp2, "%llu",           /* size quota */
                     pwinfo->quota_size) < 0) {
             return -1;
-        }        
+        }
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
     if (pwinfo->allow_local_ip != NULL) {
         fprintf(fp2, "%s", pwinfo->allow_local_ip);   /* allowed local ip */
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
     if (pwinfo->deny_local_ip != NULL) {
         fprintf(fp2, "%s", pwinfo->deny_local_ip);    /* denied local ip */
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
-    }    
+    }
     if (pwinfo->allow_client_ip != NULL) {
         fprintf(fp2, "%s", pwinfo->allow_client_ip);  /* allowed client ip */
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
     if (pwinfo->deny_client_ip != NULL) {
         fprintf(fp2, "%s", pwinfo->deny_client_ip);   /* denied local ip */
     }
-    if (fprintf(fp2, PW_LINE_SEP) < 0) {    
+    if (fprintf(fp2, PW_LINE_SEP) < 0) {
         return -1;
     }
     if (pwinfo->has_time != 0) {
@@ -699,19 +668,19 @@ static int add_new_pw_line(FILE * const fp2, const PWInfo * const pwinfo)
     if (fprintf(fp2, "\n") < 0) {
         return -1;
     }
-    
+
     return 0;
 }
 
 static char *do_get_passwd(void)
 {
     static char pwd[LINE_MAX];
-    char pwd2[LINE_MAX];    
+    char pwd2[LINE_MAX];
     int tries = MAX_PASSWD_CHANGE_TRIES;
-       
+
     *pwd = 0;
     *pwd2 = 0;
-    
+
     again:
     printf("Password: ");
     fflush(stdout);
@@ -732,24 +701,24 @@ static char *do_get_passwd(void)
     puts("");
     if (strcmp(pwd, pwd2) != 0) {
         if (*pwd2 != 0) {
-            memset(pwd2, 0, strlen(pwd2));
+            pure_memzero(pwd2, strlen(pwd2));
         }
         if (*pwd != 0) {
-            memset(pwd, 0, strlen(pwd));
+            pure_memzero(pwd, strlen(pwd));
         }
         puts("You didn't enter the same password");
         if (--tries > 0) {
             goto again;
         }
         enable_echo();
-        
+
         return NULL;
     }
     if (*pwd2 != 0) {
-        memset(pwd2, 0, strlen(pwd2));
+        pure_memzero(pwd2, strlen(pwd2));
     }
     enable_echo();
-    
+
     return pwd;
 }
 
@@ -757,15 +726,15 @@ static int do_list(const char * const file)
 {
     FILE *fp;
     PWInfo pwinfo;
-    char line[LINE_MAX];    
-    
+    char line[LINE_MAX];
+
     if (file == NULL) {
         fprintf(stderr, "missing file to list accounts\n");
         return PW_ERROR_MISSING_PASSWD_FILE;
     }
     if ((fp = fopen(file, "r")) == NULL) {
         perror("Unable to open the passwd file");
-        return PW_ERROR_MISSING_PASSWD_FILE;        
+        return PW_ERROR_MISSING_PASSWD_FILE;
     }
     while (fgets(line, (int) sizeof line - 1U, fp) != NULL) {
         strip_lf(line);
@@ -779,7 +748,7 @@ static int do_list(const char * const file)
         if (isatty(1)) {
             printf("%-19s %-39s %-19s\n", pwinfo.login, pwinfo.home, pwinfo.gecos);
         } else {
-            printf("%s\t%s\t%s\n", pwinfo.login, pwinfo.home, pwinfo.gecos);            
+            printf("%s\t%s\t%s\n", pwinfo.login, pwinfo.home, pwinfo.gecos);
         }
     }
     fclose(fp);
@@ -793,7 +762,7 @@ static int do_useradd(const char * const file,
     char *file2;
     FILE *fp2;
     PWInfo pwinfo = *pwinfo_;
-    
+
     if (pwinfo.login == NULL || *(pwinfo.login) == 0) {
         fprintf(stderr, "Missing login\n");
         return PW_ERROR_MISSING_LOGIN;
@@ -802,21 +771,23 @@ static int do_useradd(const char * const file,
         fprintf(stderr, "Missing passwd file\n");
         return PW_ERROR_MISSING_PASSWD_FILE;
     }
+#ifndef ACCEPT_ROOT_VIRTUAL_USERS
     if (pwinfo.uid <= (uid_t) 0 || pwinfo.gid <= (gid_t) 0) {
         fprintf(stderr, "You must give (non-root) uid and gid\n");
         return PW_ERROR_USERADD_NOT_ROOT;
     }
+#endif
     if (pwinfo.home == NULL) {
-        fprintf(stderr, "Missing home directory\n");        
+        fprintf(stderr, "Missing home directory\n");
         return PW_ERROR_USERADD_MISSING_HOME_DIR;
     }
     if (pwinfo.gecos == NULL) {
         if ((pwinfo.gecos = strdup("")) == NULL) {
             no_mem();
         }
-    }           
+    }
     if ((pwinfo.pwd = do_get_passwd()) == NULL) {
-        fprintf(stderr, "Error with entering password - aborting\n");        
+        fprintf(stderr, "Error with entering password - aborting\n");
         return PW_ERROR_ENTER_PASSWD_PW_ERROR;
     }
     {
@@ -824,20 +795,20 @@ static int do_useradd(const char * const file,
 
         pwinfo.pwd = best_crypt(cleartext);
         if (*cleartext != 0) {
-            memset(cleartext, 0, strlen(cleartext));
+            pure_memzero(cleartext, strlen(cleartext));
         }
-    }            
+    }
     if ((file2 = newpasswd_filename(file)) == NULL) {
         no_mem();
     }
     if ((fp2 = create_newpasswd(file, file2, pwinfo.login, 1, 0)) == NULL) {
         fprintf(stderr, "Error.\n"
                 "Check that [%s] doesn't already exist,\n"
-                "and that [%s] can be written.\n", 
+                "and that [%s] can be written.\n",
                 pwinfo.login, file2);
-        free(file2);        
+        free(file2);
         return PW_ERROR_USER_ALREADY_EXIST;
-    }    
+    }
     if (add_new_pw_line(fp2, &pwinfo) != 0) {
         fprintf(stderr, "Unable to append a line\n");
         goto bye;
@@ -845,7 +816,7 @@ static int do_useradd(const char * const file,
     fflush(fp2);
 #ifdef HAVE_FILENO
     fsync(fileno(fp2));
-#endif  
+#endif
     if (fclose(fp2) != 0) {
         perror("Unable to close the file");
         goto bye2;
@@ -856,13 +827,13 @@ static int do_useradd(const char * const file,
     }
     free(file2);
     return 0;
-    
+
     bye:
     fclose(fp2);
     bye2:
     unlink(file2);
     free(file2);
-    
+
     return PW_ERROR_UNEXPECTED_ERROR;
 }
 
@@ -893,8 +864,8 @@ static int do_usermod(const char * const file,
 
         fetched_info.pwd = best_crypt(cleartext);
         if (*cleartext != 0) {
-            memset(cleartext, 0, strlen(cleartext));
-        }        
+            pure_memzero(cleartext, strlen(cleartext));
+        }
     }
     if (pwinfo->uid > (uid_t) 0) {
         fetched_info.uid = pwinfo->uid;
@@ -918,7 +889,7 @@ static int do_usermod(const char * const file,
     }
     if (pwinfo->has_bw_ul != 0) {
         if (pwinfo->has_bw_ul < 0) {
-            fetched_info.has_bw_ul = 0;            
+            fetched_info.has_bw_ul = 0;
         } else {
             fetched_info.has_bw_ul = pwinfo->has_bw_ul;
             fetched_info.bw_ul = pwinfo->bw_ul;
@@ -934,7 +905,7 @@ static int do_usermod(const char * const file,
     }
     if (pwinfo->has_quota_size != 0) {
         if (pwinfo->has_quota_size < 0) {
-            fetched_info.has_quota_size = 0;            
+            fetched_info.has_quota_size = 0;
         } else {
             fetched_info.has_quota_size = pwinfo->has_quota_size;
             fetched_info.quota_size = pwinfo->quota_size;
@@ -942,7 +913,7 @@ static int do_usermod(const char * const file,
     }
     if (pwinfo->has_ul_ratio != 0) {
         if (pwinfo->has_ul_ratio < 0) {
-            fetched_info.has_ul_ratio = 0;            
+            fetched_info.has_ul_ratio = 0;
         } else {
             fetched_info.has_ul_ratio = pwinfo->has_ul_ratio;
             fetched_info.ul_ratio = pwinfo->ul_ratio;
@@ -979,7 +950,7 @@ static int do_usermod(const char * const file,
     }
     if (pwinfo->has_per_user_max != 0) {
         if (pwinfo->has_per_user_max < 0) {
-            fetched_info.has_per_user_max = 0;            
+            fetched_info.has_per_user_max = 0;
         } else {
             fetched_info.has_per_user_max = pwinfo->has_per_user_max;
             fetched_info.per_user_max = pwinfo->per_user_max;
@@ -994,7 +965,7 @@ static int do_usermod(const char * const file,
                 "and that [%s] can be written.\n", pwinfo->login, file2);
         free(file2);
         return PW_ERROR_USER_ALREADY_EXIST;
-    }    
+    }
     if (add_new_pw_line(fp2, &fetched_info) != 0) {
         fprintf(stderr, "Unable to append a line\n");
         goto bye;
@@ -1002,7 +973,7 @@ static int do_usermod(const char * const file,
     fflush(fp2);
 #ifdef HAVE_FILENO
     fsync(fileno(fp2));
-#endif      
+#endif
     if (fclose(fp2) != 0) {
         perror("Unable to close the file");
         goto bye2;
@@ -1013,13 +984,13 @@ static int do_usermod(const char * const file,
     }
     free(file2);
     return 0;
-    
+
     bye:
     fclose(fp2);
     bye2:
     unlink(file2);
     free(file2);
-    
+
     return PW_ERROR_UNEXPECTED_ERROR;
 }
 
@@ -1028,7 +999,7 @@ static int do_userdel(const char * const file,
 {
     char *file2;
     FILE *fp2;
-    
+
     if (pwinfo->login == NULL || *(pwinfo->login) == 0) {
         fprintf(stderr, "Missing login\n");
         return -1;
@@ -1036,7 +1007,7 @@ static int do_userdel(const char * const file,
     if (file == NULL) {
         fprintf(stderr, "Missing passwd file\n");
         return PW_ERROR_MISSING_PASSWD_FILE;
-    }    
+    }
     if ((file2 = newpasswd_filename(file)) == NULL) {
         no_mem();
     }
@@ -1061,11 +1032,11 @@ static int do_userdel(const char * const file,
     }
     free(file2);
     return 0;
-    
+
     bye2:
     unlink(file2);
     free(file2);
-    
+
     return PW_ERROR_UNEXPECTED_ERROR;
 }
 
@@ -1075,9 +1046,9 @@ static int do_show(const char * const file, const PWInfo * const pwinfo)
     struct passwd *pwd;
     struct group *grp;
     const char *pwd_name = "-";
-    const char *grp_name = "-";    
+    const char *grp_name = "-";
     static char line[LINE_MAX];
-    
+
     if (pwinfo->login == NULL || *(pwinfo->login) == 0) {
         fprintf(stderr, "Missing login\n");
         return PW_ERROR_MISSING_LOGIN;
@@ -1111,11 +1082,11 @@ static int do_show(const char * const file, const PWInfo * const pwinfo)
            "Max size           : %llu Mb (%s)\n"
            "Ratio              : %u:%u (%s:%s)\n"
            "Allowed local  IPs : %s\n"
-           "Denied  local  IPs : %s\n"           
+           "Denied  local  IPs : %s\n"
            "Allowed client IPs : %s\n"
-           "Denied  client IPs : %s\n"           
+           "Denied  client IPs : %s\n"
            "Time restrictions  : %04u-%04u (%s)\n"
-       "Max sim sessions   : %u (%s)\n"
+           "Max sim sessions   : %u (%s)\n"
            "\n",
            fetched_info.login,
            fetched_info.pwd,
@@ -1136,15 +1107,15 @@ static int do_show(const char * const file, const PWInfo * const pwinfo)
            SHOW_STATE(fetched_info.has_ul_ratio),
            SHOW_STATE(fetched_info.has_dl_ratio),
            SHOW_STRING(fetched_info.allow_local_ip),
-           SHOW_STRING(fetched_info.deny_local_ip),           
+           SHOW_STRING(fetched_info.deny_local_ip),
            SHOW_STRING(fetched_info.allow_client_ip),
            SHOW_STRING(fetched_info.deny_client_ip),
            SHOW_IFEN(fetched_info.has_time, fetched_info.time_begin),
            SHOW_IFEN(fetched_info.has_time, fetched_info.time_end),
            SHOW_STATE(fetched_info.has_time),
-       SHOW_IFEN(fetched_info.has_per_user_max, fetched_info.per_user_max),
+           SHOW_IFEN(fetched_info.has_per_user_max, fetched_info.per_user_max),
            SHOW_STATE(fetched_info.per_user_max));
-    
+
     return 0;
 }
 
@@ -1160,12 +1131,10 @@ static int do_passwd(const char * const file,
         return PW_ERROR_MISSING_PASSWD_FILE;
     }
     if ((pwinfo->pwd = do_get_passwd()) == NULL) {
-        fprintf(stderr, "Error with entering password - aborting\n");        
+        fprintf(stderr, "Error with entering password - aborting\n");
         return PW_ERROR_ENTER_PASSWD_PW_ERROR;
-    }    
-    do_usermod(file, pwinfo);
-    
-    return 0;
+    }
+    return do_usermod(file, pwinfo);
 }
 
 static int do_mkdb(const char *dbfile, const char * const file)
@@ -1179,20 +1148,20 @@ static int do_mkdb(const char *dbfile, const char * const file)
     PureDBW dbw;
     int ret = PW_ERROR_UNEXPECTED_ERROR;
     char line[LINE_MAX];
-    
+
     if (dbfile == NULL || *dbfile == 0) {
         char *dbfile_;
-        
+
         if ((dbfile_ = getenv(ENV_DEFAULT_PW_DB)) != NULL && *dbfile_ != 0) {
             dbfile = dbfile_;
-        } else {        
+        } else {
             dbfile = DEFAULT_PW_DB;
         }
     }
     if (file == NULL) {
         fprintf(stderr, "Missing passwd file\n");
         return PW_ERROR_MISSING_PASSWD_FILE;
-    }    
+    }
     if ((fp = fopen(file, "r")) == NULL) {
         perror("Unable to open the passwd file");
         return PW_ERROR_MISSING_PASSWD_FILE;
@@ -1239,9 +1208,9 @@ static int do_mkdb(const char *dbfile, const char * const file)
     err:
     puredbw_free(&dbw);
     ALLOCA_FREE(index_dbfile);
-    ALLOCA_FREE(data_dbfile);    
+    ALLOCA_FREE(data_dbfile);
     fclose(fp);
-    
+
     return ret;
 }
 
@@ -1249,14 +1218,16 @@ static void init_zrand(void)
 {
     struct timeval tv;
     struct timezone tz;
-    
+
     gettimeofday(&tv, &tz);
 #ifdef HAVE_SRANDOMDEV
     srandomdev();
 #elif defined(HAVE_RANDOM)
-    srandom((unsigned int) (tv.tv_sec ^ tv.tv_usec ^ (getpid() << 16)));
+    srandom((unsigned int) (tv.tv_sec ^ tv.tv_usec ^
+                            ((unsigned int) getpid() << 16)));
 #else
-    srand((unsigned int) (tv.tv_sec ^ tv.tv_usec ^ (getpid() << 16)));
+    srand((unsigned int) (tv.tv_sec ^ tv.tv_usec ^
+                          ((unsigned int) getpid() << 16)));
 #endif
 }
 
@@ -1270,7 +1241,7 @@ int main(int argc, char *argv[])
     int ret = 0;
     int with_chroot = 1;
     int with_mkdb = 0;
-        
+
     if (argc < 2) {
         help();
     }
@@ -1285,17 +1256,14 @@ int main(int argc, char *argv[])
 # ifdef LC_COLLATE
     (void) setlocale(LC_COLLATE, "");
 # endif
-#endif    
-
-#ifdef PROBE_RANDOM_AT_RUNTIME
-    pw_zrand_probe();
 #endif
-    
+
     pwinfo.pwd = NULL;
     pwinfo.gecos = NULL;
     pwinfo.home = NULL;
     pwinfo.allow_local_ip = pwinfo.deny_local_ip = NULL;
-    pwinfo.allow_client_ip = pwinfo.deny_client_ip = NULL;        
+    pwinfo.allow_client_ip = pwinfo.deny_client_ip = NULL;
+    pwinfo.bw_dl = pwinfo.bw_ul = 0UL;
     pwinfo.has_bw_dl = 0;
     pwinfo.has_bw_ul = 0;
     pwinfo.has_quota_files = 0;
@@ -1313,7 +1281,7 @@ int main(int argc, char *argv[])
     pwinfo.uid = (uid_t) 0U;
     pwinfo.gid = (gid_t) 0U;
 //#endif
-    
+
     argv++;
     argc--;
     action = *argv;
@@ -1326,8 +1294,8 @@ int main(int argc, char *argv[])
     }
     filter_pw_line_sep(pwinfo.login);
     while ((fodder =
-	    getopt(argc, argv, 
-		   "c:d:D:f:F:g:hi:I:mn:N:q:Q:r:R:t:T:u:y:z:")) != -1) {
+            getopt(argc, argv,
+                   "c:d:D:f:F:g:hi:I:mn:N:q:Q:r:R:t:T:u:y:z:")) != -1) {
         switch(fodder) {
         case 'c' : {
             if ((pwinfo.gecos = strdup(optarg)) == NULL) {
@@ -1342,7 +1310,7 @@ int main(int argc, char *argv[])
             char *optarg_copy;
             size_t sizeof_home;
             size_t optarg_len;
-            
+
             if ((optarg_copy = strdup(optarg)) == NULL) {
                 no_mem();
             }
@@ -1363,7 +1331,7 @@ int main(int argc, char *argv[])
             }
             snprintf(pwinfo.home, sizeof_home, "%s%s", optarg_copy,
                      with_chroot != 0 ? "/./" : "");
-            filter_pw_line_sep(pwinfo.home);            
+            filter_pw_line_sep(pwinfo.home);
             break;
         }
         case 'f' : {
@@ -1380,16 +1348,16 @@ int main(int argc, char *argv[])
         }
         case 'g' : {
             struct group *gr;
-            
+
             if (pwinfo.gid > (gid_t) 0 && pwinfo.uid <= (uid_t) 0) {
                 fprintf(stderr, "You already gave a gid\n");
                 exit(EXIT_FAILURE);
-            }                
+            }
             if ((gr = getgrnam(optarg)) != NULL) {
                 pwinfo.gid = gr->gr_gid;
             } else {
                 pwinfo.gid = (gid_t) strtoul(optarg, NULL, 10);
-            }            
+            }
             break;
         }
         case 'h' : {
@@ -1425,7 +1393,7 @@ int main(int argc, char *argv[])
             if (*optarg == 0) {
                 pwinfo.has_quota_size = -1;
             } else {
-                pwinfo.quota_size = strtoull(optarg, NULL, 10) * 
+                pwinfo.quota_size = strtoull(optarg, NULL, 10) *
                     (1024ULL * 1024ULL);
                 pwinfo.has_quota_size = 1;
             }
@@ -1443,7 +1411,7 @@ int main(int argc, char *argv[])
                 pwinfo.has_ul_ratio = 1;
             }
             break;
-        }            
+        }
         case 'Q' : {
             if (*optarg == 0) {
                 pwinfo.has_dl_ratio = -1;
@@ -1452,7 +1420,7 @@ int main(int argc, char *argv[])
                 if (pwinfo.dl_ratio < 1U) {
                     fprintf(stderr, "Illegal download ratio\n");
                     exit(EXIT_FAILURE);
-                }            
+                }
                 pwinfo.has_dl_ratio = 1;
             }
             break;
@@ -1468,13 +1436,13 @@ int main(int argc, char *argv[])
                 no_mem();
             }
             break;
-        }            
+        }
         case 't' : {
             if (*optarg == 0) {
                 pwinfo.has_bw_dl = -1;
             } else {
                 if ((pwinfo.bw_dl = strtoul(optarg, NULL, 10)) > 0UL) {
-                    pwinfo.bw_dl *= 1024UL;                    
+                    pwinfo.bw_dl *= 1024UL;
                     pwinfo.has_bw_dl = 1;
                 }
             }
@@ -1490,10 +1458,10 @@ int main(int argc, char *argv[])
                 }
             }
             break;
-        }            
+        }
         case 'u' : {
-            struct passwd *pw;                
-            
+            struct passwd *pw;
+
             if (pwinfo.uid > (uid_t) 0) {
                 fprintf(stderr, "You already gave an uid\n");
                 exit(EXIT_FAILURE);
@@ -1508,16 +1476,16 @@ int main(int argc, char *argv[])
             }
             break;
         }
-    case 'y' : {
-        if ((pwinfo.per_user_max = (unsigned int) strtoul(optarg, NULL, 10)) <= 0U) {
+        case 'y' : {
+            if ((pwinfo.per_user_max = (unsigned int) strtoul(optarg, NULL, 10)) <= 0U) {
                 pwinfo.has_per_user_max = -1;
             } else {
                 pwinfo.has_per_user_max = 1;
             }
-        break;
-    }
+            break;
+        }
         case 'z' : {
-            if (sscanf(optarg, "%u-%u", 
+            if (sscanf(optarg, "%u-%u",
                        &pwinfo.time_begin, &pwinfo.time_end) == 2 &&
                 pwinfo.time_begin < 2360 && (pwinfo.time_begin % 100) < 60 &&
                 pwinfo.time_end < 2360 && (pwinfo.time_end % 100) < 60) {
@@ -1525,7 +1493,7 @@ int main(int argc, char *argv[])
             } else if (*optarg != 0) {
                 fprintf(stderr, "Time should be given as hhmm-hhmm\n"
                         "Example : 0900-1800 (9 am to 6 pm)\n");
-                exit(EXIT_FAILURE);                    
+                exit(EXIT_FAILURE);
             } else {
                 pwinfo.has_time = -1;
             }
@@ -1537,7 +1505,7 @@ int main(int argc, char *argv[])
     }
     if (file == NULL) {
         char *file_;
-        
+
         if ((file_ = getenv(ENV_DEFAULT_PW_FILE)) != NULL && *file_ != 0) {
             file = file_;
         } else if ((file = strdup(DEFAULT_PW_FILE)) == NULL) {
@@ -1555,17 +1523,17 @@ int main(int argc, char *argv[])
         ret = do_usermod(file, &pwinfo);
         if (with_mkdb != 0) {
             ret |= do_mkdb(dbfile, file);
-        }        
+        }
     } else if (strcasecmp(action, "userdel") == 0) {
         ret = do_userdel(file, &pwinfo);
         if (with_mkdb != 0) {
             ret |= do_mkdb(dbfile, file);
-        }        
+        }
     } else if (strcasecmp(action, "passwd") == 0) {
-        do_passwd(file, &pwinfo);
+        ret = do_passwd(file, &pwinfo);
         if (with_mkdb != 0) {
-            do_mkdb(dbfile, file);
-        }        
+            ret |= do_mkdb(dbfile, file);
+        }
     } else if (strcasecmp(action, "show") == 0) {
         ret = do_show(file, &pwinfo);
     } else if (strcasecmp(action, "mkdb") == 0) {
@@ -1576,7 +1544,8 @@ int main(int argc, char *argv[])
         ret = PW_ERROR_UNEXPECTED_ERROR;
         help();
     }
-               
+    free(pwinfo.gecos);
+
     return ret;
 }
 
